import logging

FORMAT = '%(levelname)s: %(message)s'
logging.basicConfig(format=FORMAT)


def setup_logger(name, lvl='INFO'):
    logger = logging.getLogger(name)
    logger.setLevel(getattr(logging, lvl))
    return logger


logger = setup_logger('Feature Selector')



import numpy as np
from multiset_multicover import GreedyCoverInstance
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_is_fitted

from ._logger import logger


class GCIWrapper(BaseEstimator):
    def __init__(self, n_elements, multiplier=None, verbose=True):
        """
        Constructs a wrapper for a GreedyCoverInstance that includes
        different functionality such as plotting.

        Args:
            n_elements: int, number of elements in the greey cover
            multiplier: numeric or None. If not None, will multiply each
                element of M with this value, effectively increasing the
                multiplicity of each value. This can be applied when finer
                resolution over the multiplities is needed in case when
                they are represented as float values. Note, the coverage
                factor requested will also be multiplied by this same factor.
            verbose: bool, whether to show info.

        Attributes:
            n_elements: int, total number of elements, i.e., rows of M
            n_multisets_: int, number of multisets actually used, i.e.
                columns with a single nonzero value in M
            max_coverage_: ndarray, maximum possible coverage for every
                element
            multisets_incomplete_cover_: ndarray, indices of multisets
                that could not be covered to the desired coverage
            coverage_until_: ndarray, coverage obtained after every
                iteration. Has shape (solution.size,).
            n_elements_remaining_: ndarray, number of elements remaining
                after every iteration. Has shape (solution.size,)

        """
        if multiplier is not None and multiplier < 0:
            raise ValueError("Cannot have a negative multiplier.")
        self.n_elements = n_elements
        self._gci = GreedyCoverInstance(n_elements)
        self.multiplier = multiplier
        self.verbose = verbose

    def __getitem__(self, i):
        """
        Returns a pair of arrays corresponding to the elements and
        multiplicities of feature i.
        """
        # First convert to the inner index
        i = self._real_index_to_inner(i)
        if i is None:
            return np.array([], dtype=int), np.array([])
        elements, multiplicity = self._gci.__getitem__(i)
        elements = np.array(elements)
        multiplicity = self._divide_for_multiplier(multiplicity)
        return elements, multiplicity

    def effective_at(self, i):
        """
        Returns a pair of arrays corresponding to the elements and
        effective multiplicities of feature i. The effective multiplicity
        of a feature is the actual multiplicities that were needed
        from that feature at the moment of selection. If the feature was
        not selected, then this will return the multiplicities that remained
        for that feature at the end of the cover.
        """
        i = self._real_index_to_inner(i)
        if i is None:
            return np.array([], dtype=int), np.array([])
        elements, multiplicity = self._gci.effective_at(i)
        elements = np.array(elements)
        multiplicity = self._divide_for_multiplier(multiplicity)
        return elements, multiplicity

    def fit(self, M):
        """
        Given a matrix M, construct a greedy cover instance by considering
        the columns of M as multisets. Will skip all-zero columns and
        save the "true" indices of the columns under ``self._true_indices``.

        Args:
            M: np.ndarray, the data matrix. Has shape (n_elements, n_multisets)
        """
        M = self._multiply_for_multiplier(M)
        M = np.round(M).astype(int)

        # Determine features with non-zero entries and push them as multisets.
        self._true_indices = []
        for i, col in enumerate(M.T):
            elements = col.nonzero()[0]
            if elements.size > 0:
                multiplicity = col[elements]
                self._true_indices.append(i)
                self._gci.add_multiset(
                    elements.tolist(), multiplicity.tolist())

        self._true_indices = np.array(self._true_indices)
        self.n_multisets_ = len(self._true_indices)
        if self.verbose:
            logger.info(f"Saved {self.n_multisets_}/{M.shape[1]} multisets.")

    def predict(self, coverage, max_iters=0):
        """
        Returns the indices of the selected features given
        a certain coverage.

        Args:
            coverage: int or list, in case of a list will apply a specific
                coverage to each element.
            max_iters: maximum number of iterations (features) to return.
                A value of 0 or None means no limit.
        """
        check_is_fitted(self)
        coverage = self._multiply_for_multiplier(coverage)
        # Need to convert to Pythonic list in this case, since
        # GreedyCoverInstance instance only accepts python lists.
        if isinstance(coverage, np.ndarray):
            coverage = coverage.tolist()
        # Do not forget to convert back to the true indices
        return self._true_indices[self._gci.cover(coverage, max_iters)]

    def fit_predict(self, M, coverage, max_iters=0):
        """
        Given a matrix M, construct a greedy cover instance by considering
        the columns of M as multisets. Will skip all-zero columns and
        save the "true" indices of the columns under ``self._true_indices``.
        Then returns the indices of the selected features given
        a certain coverage.

        Args:
            M: np.ndarray, the data matrix. Has shape (n_elements, n_multisets)
            coverage: int or list, in case of a list will apply a specific
                coverage to each element.
            max_iters: maximum number of iterations (features) to return.
                A value of 0 or None means no limit.
        """
        self.fit(M)
        return self.predict(coverage, max_iters)

    @property
    def max_coverage_(self):
        """
        The total multiplicity of each element across all multisets.
        """
        return self._divide_for_multiplier(self._gci.max_coverage_)

    @property
    def elements_incomplete_cover_(self):
        """
        Array of indices corresponding to elements that could not
        be covered to the desired coverage.
        """
        return np.array(self._gci.multisets_incomplete_cover_)

    @property
    def coverage_until_(self):
        """
        Array of coverage factors up until and including every
        multiset selected. Has shape (n_solution,)
        """
        return self._divide_for_multiplier(self._gci.coverage_until_)

    @property
    def n_elements_remaining_(self):
        """
        Array of integers corresponding to the number of elements
        that remain up until a multiset has been selected.
        Has shape (n_solution,).
        """
        return np.array(self._gci.n_elements_remaining_)

    def _real_index_to_inner(self, i):
        """
        Converts a real index to an inner index. If the real index was
        not used to construct GCI, returns None.
        """
        check_is_fitted(self)
        if i in self._true_indices:
            return (self._true_indices == i).nonzero()[0][0]
        return None

    def _divide_for_multiplier(self, array):
        if not isinstance(array, np.ndarray):
            array = np.array(array)
        array = array.astype(float).copy()
        if self.multiplier is not None:
            array /= self.multiplier
        return array

    def _multiply_for_multiplier(self, array):
        if not isinstance(array, np.ndarray):
            array = np.array(array)
        array = array.copy()
        if self.multiplier is not None:
            array *= self.multiplier
        return array


class GCIPython(BaseEstimator):
    def __init__(self, verbose=True, **kwargs):
        """
        Constructs a wrapper for a GreedyCoverInstance based on a Python
        implementation. Results are identical to GCIWrapper above.
        Use mainly for debugging purposes.

        Args:
            verbose: bool, whether to show info.

        Attributes:
            n_elements: int, total number of elements, i.e., rows of M
            n_multisets_: int, number of multisets actually used, i.e.
                columns with a single nonzero value in M
            max_coverage_: ndarray, maximum possible coverage for every
                element
            multisets_incomplete_cover_: ndarray, indices of multisets
                that could not be covered to the desired coverage
            coverage_until_: ndarray, coverage obtained after every
                iteration. Has shape (solution.size,).
            n_elements_remaining_: ndarray, number of elements remaining
                after every iteration. Has shape (solution.size,)

        """
        self.verbose = verbose

    def __getitem__(self, i):
        check_is_fitted(self)
        if i in self._nonzero_cols:
            return self._M_trimmed[np.where(self._nonzero_cols == i)[0][0]]
        return np.zeros(self.n_elements_.size, dtype=self._M_trimmed.dtype)

    def effective_at(self, i):
        check_is_fitted(self)
        if not hasattr(self, 'effective'):
            raise ValueError("No coverage has been specified.")

        i = self._real_index_to_inner(i)
        if i is None or i not in self._solution:
            return np.zeros(
                self.effective_[0].size, dtype=self.effective_.dtype)

        index = np.where(self._solution == i)[0][0]
        return self.effective_[index]

    def fit(self, M, multiplier=None):
        """
        M: array, rows are elements, columns will be used to construct
            multisets.
        multiplier: numeric or None. If not None, will multiply each
            element of M with this value, effectively increasing the
            multiplicity of each value. This can be applied when finer
            resolution over the multiplities is needed in case when
            they are represented as float values. Note, the coverage
            factor requested will also be multiplied by this same factor.
        """
        if multiplier is not None and multiplier < 0:
            raise ValueError("Cannot have a negative multiplier.")
        self.multiplier_ = multiplier
        self.n_elements = M.shape[0]

        M = self._correct_x_multiplier(M)
        M = np.round(M)

        if M.min() < 0:
            raise ValueError("M cannot contain negative values.")

        self._nonzero_cols = np.where(~np.all(M == 0, axis=0))[0]
        self._M_trimmed = M[:, self._nonzero_cols]
        self.n_multisets_ = len(self._nonzero_cols)
        if self.verbose:
            logger.info(f"Saved {self.n_multisets_}/{M.shape[1]} multisets.")

    @property
    def max_coverage_(self):
        """
        Maximum number of times an element can be covered.
        """
        check_is_fitted(self)
        return self._M_trimmed.sum(axis=1)

    def predict(self, coverage, max_iters=0):
        """
        Returns the indices of the selected features given
        a certain coverage.

        Args:
            coverage: int or list, in case of a list will apply a specific
                coverage to each element.
            max_iters: maximum number of iterations (features) to return.
                A value of 0 or None means no limit.
        """
        self.coverage = coverage
        check_is_fitted(self)
        coverage = self._correct_x_multiplier(coverage)
        self.effective_coverage_ = coverage
        self.max_iters = max_iters

        def stop(x, iter, to_cover):
            if max_iters > 0:
                if iter >= max_iters:
                    logger.info("Reached maximum number of iterations.")
                    return True
            if x.max() == 0:
                logger.info("No more nonzero multisets remain.")
                return True
            if to_cover.max() == 0:
                logger.info("Finished cover.")
                return True
            return False

        temp_M = np.ascontiguousarray(self._M_trimmed.copy())
        temp_M = np.clip(temp_M, 0, coverage)

        if isinstance(coverage, int):
            to_cover = np.full(temp_M.shape[0], coverage, dtype=temp_M.dtype)
        else:
            to_cover = coverage.copy()
        to_cover = np.minimum(to_cover, self.max_coverage_)

        self._solution = []
        self.effective_ = []
        self.coverage_until_ = []
        self.n_elements_remaining_ = []
        iter = 0
        __total_col_sum = self._M_trimmed.sum(axis=0)

        while not stop(temp_M, iter, to_cover):
            # Find multiset with highest value.
            # In case of a collision, take the multiset with the highest
            # total value as determined by _M_trimmed.
            __col_sum = temp_M.sum(axis=0)
            __top_val = np.max(__col_sum)
            __achieve_top_val = np.argwhere(__col_sum == __top_val).flatten()
            if __achieve_top_val.size == 1:
                next_mset = __achieve_top_val.item()
            else:
                __best_total_val = np.argmax(__total_col_sum[__achieve_top_val])
                next_mset = __achieve_top_val[__best_total_val]

            cur_mset_vals = temp_M[:, next_mset].copy()
            to_cover = np.maximum(0, to_cover - cur_mset_vals)
            if isinstance(coverage, np.ndarray):
                self.coverage_until_.append(np.min(coverage - to_cover))
            else:
                self.coverage_until_.append(coverage - to_cover.max())
            self.n_elements_remaining_.append(np.sum(to_cover != 0))

            self.effective_.append(cur_mset_vals)
            self._solution.append(next_mset)
            # Correct the multiplicity matrix by subtracting
            # the selected multiset
            temp_M = np.minimum(temp_M, to_cover.reshape(-1, 1))
            # Remove this multiset
            temp_M[:, next_mset] = 0
            iter += 1

        self._solution = np.array(self._solution)
        self.coverage_until_ = np.array(self.coverage_until_)
        self.n_elements_remaining_ = np.array(self.n_elements_remaining_)
        self.effective_ = np.array(self.effective_)

        solution = self._nonzero_cols[np.array(self._solution)]

        return solution.copy()

    def fit_predict(self, M, coverage, max_iters=0, multiplier=None):
        """
        Given a matrix M, construct a greedy cover instance by considering
        the columns of M as multisets. Will skip all-zero columns and
        save the "true" indices of the columns under ``self._true_indices``.
        Then returns the indices of the selected features given
        a certain coverage.

        Args:
            M: np.ndarray, the data matrix. Has shape (n_elements, n_multisets)
            coverage: int or list, in case of a list will apply a specific
                coverage to each element.
            max_iters: maximum number of iterations (features) to return.
                A value of 0 or None means no limit.
        """
        self.fit(M, multiplier)
        return self.predict(coverage, max_iters)

    @property
    def elements_incomplete_cover_(self):
        """
        Array of indices corresponding to elements that could not
        be covered to the desired coverage.
        """
        return np.where((self.max_coverage_ - self.effective_coverage_) < 0)[0]

    def _real_index_to_inner(self, i):
        """
        Converts a real index to an inner index. If the real index was
        not used to construct GCI, returns None.
        """
        check_is_fitted(self)
        if i in self._nonzero_cols:
            return (self._nonzero_cols == i).nonzero()[0][0]
        return None

    def _correct_x_multiplier(self, val):
        if isinstance(val, list):
            val = np.array(val)
        elif isinstance(val, np.ndarray):
            val = val.copy()
        if self.multiplier_ is not None:
            return val * self.multiplier_
        return val



from abc import abstractmethod

from sklearn.base import BaseEstimator


class FeatureSelector(BaseEstimator):
    """Base class of feature selector methods.

    This class should not be used directly.
    """
    @abstractmethod
    def fit(self, X, y=None, **fit_params):
        """Fit the model with X and y.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where `n_samples` is the number of samples and
            `n_features` is the number of features.
        y: array-like of shape (n_samples,)

        Returns
        -------
        self : object
            Returns the instance itself.
        """

    @abstractmethod
    def select(self, **select_params):
        """Performs feature selection.

        Returns
        -------
        self : object
            Returns the instance itself.
        """

    def fit_select(self, X, y=None, **params):
        """Fit the model with X and y and perform feature selection.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where `n_samples` is the number of samples and
            `n_features` is the number of features.
        y: array-like of shape (n_samples,)
        params: dictionary
            Parameters to be used for fit and select steps.

        Returns
        -------
        self.select
        """
        fit_params = {
            key: params[key]
            for key in self._get_fit_params()
            if key in params
        }
        select_params = {
            key: params[key]
            for key in self._get_select_params()
            if key in params
        }

        self.fit(X, y, **fit_params)
        return self.select(**select_params)

    def get_scores(self):
        if hasattr(self, 'feature_importances_'):
            return self.feature_importances_
        elif hasattr(self, 'coef_'):
            return self.coef_
        else:
            raise ValueError("No scores found in estimator.")

    def _get_fit_params(self):
        """Returns a list of parameter names used during `fit`.
        """
        return []

    def _get_select_params(self):
        """Returns a list of parameter names used during `select`.
        """
        return []



from itertools import combinations

import numpy as np
from sklearn.utils.validation import (as_float_array, assert_all_finite,
                                      check_X_y, column_or_1d, indexable)


def group_by(X, y, *, category_orders=None, operation=lambda x: x.mean(axis=0)):
    """Groups the samples in X by labels in y and applies `operation`
    to the aggregated groups.

    Parameters
    __________
    X: array-like of shape (n_samples, n_features)
        The data matrix.
    y: array-like of shape (n_samples,)
        The class labels.
    category_orders: array-like of shape (np.unique(y).size,)
        Order of class labels to use when constructing the matrix.
        If None, will sort the class labels alphabetically.
    operation: callable
        The function to apply to the aggregated groups.
    """
    X, y = check_X_y(X, y, accept_sparse=["csr"])
    X = indexable(X)[0]

    if category_orders is None:
        category_orders = np.unique(y)
    elif not set(category_orders).issubset(set(y)):
        # To avoid getting nan values
        raise ValueError("Found categories not present in `y`.")
    else:
        category_orders = column_or_1d(category_orders)

    if not callable(operation):
        raise ValueError("Please pass a callable operation.")

    M = np.zeros((len(category_orders), X.shape[1]))

    for i, category in enumerate(category_orders):
        _agg_values = operation(X[y == category])
        _agg_values = as_float_array(_agg_values).flatten()
        if len(_agg_values) != X.shape[1]:
            raise ValueError(
                "Operation must return a vector of size X.shape[1]"
                f"but instead found vector of size {len(_agg_values)}."
            )
        assert_all_finite(_agg_values)
        M[i] = _agg_values

    return M


def pairwise_differences(
        X, y,
        *,
        classes=None,
        ordered=False,
        operation=lambda x: x.mean(axis=0)):
    """
    Given an data matrix X, if ordered is False, construct a matrix P of shape
    (n * (n-1) / 2, X.shape[1]) where n is the number of classes in y.
    The (i*j, g) entry of P corresponds to the average expression of feature g
    in group i - average expression of feature g in group j, in absolute value.
    If ordered is True, the shape of P will be (n * (n-1), X.shape[1]) and
    the pairwise distances will be clipped at 0.

    Returns P and a dictionary of mappings: label, label -> index.

    Parameters
    _________
    X: np.ndarray of shape (n_samples, n_features)
    y: np.ndarray of shape (n_samples,)
    classes: np.ndarray or None, unique class labels in y
    ordered: bool, if True will construct a matrix of ordered
        pairwise differences. In this case the shape of P is
        (n * (n-1), X.shape[1]).
    operation: callable, operation to use when constructing the class vector.
    """
    if classes is None:
        classes = np.unique(y)

    n_classes = len(classes)
    # All pairwise combinations
    n_class_pairs = n_classes * (n_classes - 1) // 2

    # Cache the average vector of each class
    class_averages = group_by(
        X, y, category_orders=classes, operation=operation)

    # Compute the actual pairwise differences
    P = np.zeros((n_class_pairs * (1 if not ordered else 2), X.shape[1]))
    index_to_pair_dict = {}

    # Make sure to use range(n_classes) when indexing instead of classes,
    # to allow for arbitrary class labels.
    for index, (i, j) in enumerate(combinations(range(n_classes), 2)):
        difference = class_averages[i] - class_averages[j]
        if ordered:
            # Clip negative values to 0
            # Assign i - j to index and j - i to index + n_class_pairs
            P[index] = np.clip(difference, 0, None)
            index_to_pair_dict[index] = (i, j)
            P[index + n_class_pairs] = np.clip(-difference, 0, None)
            index_to_pair_dict[index + n_class_pairs] = (j, i)
        else:
            P[index] = np.abs(difference)
            index_to_pair_dict[index] = (i, j)

    return P, index_to_pair_dict



from abc import abstractmethod

import numpy as np
import matplotlib.pyplot as plt
from sklearn.utils.validation import check_is_fitted

from ._base import FeatureSelector
from ._gci_wrapper import GCIPython, GCIWrapper
from ._logger import logger
from ._operations import pairwise_differences


class SetCoverSelectorBase(FeatureSelector):
    """
    Constructs a base class for set cover selector algorithms.
    """

    def __init__(
            self,
            *,
            ordered=True,
            verbose=True,
            operation=lambda x: x.mean(axis=0)):
        self.ordered = ordered
        self.verbose = verbose
        self.operation = operation

    def fit(self, X, y):
        """
        Given a matrix X and a class label vector y, construct a matrix of
        class pairwise differences for every gene and use it to initialize
        a GreedyCoverInstance via a GCIWrapper.

        Parameters
        __________
        X: ndarray, data matrix, shape (n_samples, n_features)
        y: ndarray, class label vector, shape (n_samples,)
        multiplier: int or None, if not None will be used to multiply
            the pairwise differences to allow for finer resolution
            of the element multiplicities in the greedy cover algorithm.
        Mpath: None or str
            Path to a precomputed pairwise matrix.
        """
        tags = self._get_tags()
        X, y = self._validate_data(
            X,
            y,
            accept_sparse="csr",
            ensure_min_samples=2,
            ensure_min_features=2,
            force_all_finite=not tags.get("allow_nan", True),
        )

        self.n_samples_, self.n_features_ = X.shape
        self.classes_ = np.unique(y)
        self.n_classes_ = len(self.classes_)
        if self.n_classes_ <= 1:
            raise ValueError("At least 2 class labels must be given.")

        self.n_class_pairs_ = self.n_classes_ * (self.n_classes_ - 1)
        if not self.ordered:
            self.n_class_pairs_ //= 2

        # M_ has shape (n_class_pairs, n_features_)
        self.M_, self.index_to_pair_ = pairwise_differences(
            X, y, classes=self.classes_, ordered=self.ordered,
            operation=self.operation)

        return self

    @abstractmethod
    def select(self):
        raise NotImplementedError

    def _more_tags(self):
        return {
            "allow_nan": False,
            "requires_y": True,
        }

    def _get_fit_params(self):
        return ['multiplier']

    def _get_select_params(self):
        return ['coverage', 'max_iters']


class GreedyPC(SetCoverSelectorBase):
    """Constructs a feature selector based on the greedy cover algorithm.

    Given a data matrix X and a vector of class labels y, select
    a subset of features in X, such that the classes are as `far` from
    each other as possible w.r.t these features.

    E.g.
    ```
    >>> gcs = GreedyPC(ordered=True)
    >>> gcs.fit(X, y, multiplier=10)
    >>> transformed_X = gcs.transform(X, coverage=10)
    ```

    Parameters
    __________
    ordered: bool, if True will construct the a pairwise matrix
        of ordered pairs
    verbose: bool, whether to show info.
    use_python: bool, if True will switch to a Python implementation
        of the greedy cover algorithm. This is here just for
        debugging purposes as the C++ implementation is much faster.
    operation: callable, operation to use when constructing the class vector.

    Attributes
    __________
    n_samples_: int
        Number of samples seen during ``fit``
    n_features_: int
        Number of features seen during ``fit``
    classes_: ndarray of shape (n_classes,)
        Classes seen in the class label vector during ``fit``
    n_classes_: int
        Number of classes seen in the class label vector during ``fit``
    n_class_pairs_: int
        Number of pairs of classes that get constructed. This
        equals nchoosek(n_classes_, 2) if ordered is False,
        else n_classes_ * (n_classes - 1)
    M_: ndarray of shape (n_class_pairs_, n_features_)
        The pairwise difference matrix
    index, index_to_pair_: dict
        Dictionary that maps row indices in M_ to a pair of
        (ordered) classes
    multiplier: int
        Multiplier used to get a finer resolution when constructing
        multiset multiplicities
    coverage: int or ndarray of shape (n_class_pairs_)
        Coverage requested during ``predict``
    max_iters: int
        Maximum number of features to select during ``predict``
    n_outputs_: int
        Number of features selected during ``predict``
    n_pairs_with_incomplete_cover_: int
        Number of pairs (elements) that could not be covered
        to the desired coverage during ``predict``
    """

    def __init__(
            self,
            *,
            ordered=True,
            verbose=True,
            use_python=False,
            multiplier=None,
            operation=lambda x: x.mean(axis=0)):
        super().__init__(ordered=ordered, verbose=verbose, operation=operation)
        self.use_python = use_python
        self._multiplier = multiplier

    def select(self, coverage, *, max_iters=0):
        """
        Returns the indices of the selected features given a certain coverage.

        Parameters
        __________
        coverage: int or list, in case of a list will apply a specific
            coverage to each element.
        max_iters: maximum number of iterations (features) to return.
            A value of 0 or None means no limit.
        """
        check_is_fitted(self)
        self.multiplier = self._multiplier

        self.coverage = coverage
        self.max_iters = max_iters
        # max_iters == 0 means no limit on the number of iterations
        if max_iters is None or max_iters < 0:
            max_iters = 0

        solution = self._gci_wrapper.predict(coverage, max_iters)
        self.solution = solution
        self.feature_importances_ = np.zeros(self.n_features_)
        self.feature_importances_[solution] = np.arange(
            len(solution) + 1, 1, -1)
        if self.n_pairs_with_incomplete_cover_ > 0 and self.verbose:
            logger.warn("Could not cover "
                        f"{self.n_pairs_with_incomplete_cover_} elements.")

        self.n_outputs_ = len(solution)
        logger.info(f"Selected {self.n_outputs_} multisets.")

        return self.solution

    def plot_progress(self):
        """
        Plots the number of remaining elements to be covered,
        and the coverage reached for every feature selected.
        """
        fig, ax1 = plt.subplots()
        ax1.plot(self._gci_wrapper.n_elements_remaining_)
        ax1.set_ylabel('N remaining elements')
        ax1.set_xlabel('N features')
        ax1.tick_params(axis='y', labelcolor='blue')

        ax2 = ax1.twinx()
        coverage_until = self._gci_wrapper.coverage_until_
        ax2.plot(coverage_until, color='red')
        ax2.set_ylabel('Coverage')
        ax2.tick_params(axis='y', labelcolor='red')

        fig.tight_layout()
        return ax1

    @property
    def n_elements_remaining_per_iter_(self):
        return self._gci_wrapper.n_elements_remaining_

    @property
    def coverage_per_iter_(self):
        return self._gci_wrapper.coverage_until_

    def feature_coverage(self, index):
        """
        Returns a heatmap of pair coverages for the feature given at index.

        Parameters
        __________
        index: int, feature index in X
        effective: bool, if True will plot the corrected coverage
            of the feature, i.e., after stronger features may have been
            selected
        """
        check_is_fitted(self)

        elements, multiplicity = self._gci_wrapper[index]
        heatmap = np.zeros((self.n_classes_, self.n_classes_))

        for element, mult in zip(elements, multiplicity):
            i1, i2 = self.index_to_pair_[element]
            heatmap[i1][i2] = mult

        return heatmap

    def max_coverage(self):
        """
        Returns a heatmap of max pair coverages.
        """
        check_is_fitted(self)
        heatmap = np.zeros((self.n_classes_, self.n_classes_))
        max_coverage = self._gci_wrapper.max_coverage_

        for key in self.index_to_pair_:
            p1, p2 = self.index_to_pair_[key]
            heatmap[p1][p2] = max_coverage[key]

        return heatmap

    @property
    def n_pairs_with_incomplete_cover_(self):
        check_is_fitted(self)
        return len(self._gci_wrapper.elements_incomplete_cover_)

    @property
    def pairs_with_incomplete_cover_(self):
        """
        Returns pairs of classes that were not covered fully by the
        algorithm. Uses the original class labels.
        """
        check_is_fitted(self)
        int_pairs = [self.index_to_pair_[m]
                     for m in self._gci_wrapper.elements_incomplete_cover_]
        return [(self.classes_[i], self.classes_[j]) for i, j in int_pairs]

    @property
    def multiplier(self):
        return self._multiplier

    @multiplier.setter
    def multiplier(self, m):
        """
        Triggers a change on GCIWrapper everytime multiplier is changed
        by re-calling ``fit`` and re-constructing the multisets from M.
        """
        if m is not None and m <= 0:
            raise ValueError("Multiplier must be positive.")

        self._multiplier = m
        if self.use_python:
            logger.info("Using Python")
            self._gci_wrapper = GCIPython(self.verbose)
        else:
            self._gci_wrapper = GCIWrapper(
                self.M_.shape[0], self._multiplier, self.verbose)
        self._gci_wrapper.fit(self.M_)


class CEMPC(SetCoverSelectorBase):
    """Set cover based on the cross entropy method.

    Parameters
    __________
    ordered: bool, if True will construct the a pairwise matrix
        of ordered pairs
    verbose: bool, whether to show info.
    max_iters: int
        Maximum number of iterations for the cross entropy method.
    N: int
        Number of random samples to draw per iteration.
    rho: float in (0, 1)
        quantile to use when selecting gamma_hat.
    alpha: float:
        Parameter that controls the tradeoff between coverage desired
        and the number of features selected. If None, will use
        alpha = coverage * 10 / n_features_
    smoothing_parameter: float
        Parameter to use during exponential smoothing update. If None,
        will not apply smoothing.
    eps: float
        Tolerance to use for stopping the algorithm.
    patience: int
        Number of iterations to wait before stopping the algorithm.
    operation: callable, operation to use when constructing the class vector.
    """

    def __init__(
            self,
            *,
            ordered=True,
            verbose=True,
            max_iters=500,
            rs=1000,
            rho=0.1,
            alpha=None,
            smoothing_parameter=0.7,
            eps=1e-4,
            patience=5,
            operation=lambda x: x.mean(axis=0)):
        super().__init__(ordered=ordered, verbose=verbose, operation=operation)
        self.max_iters = max_iters
        self.rs = rs
        self.rho = rho
        self.alpha = alpha
        self.smoothing_parameter = smoothing_parameter
        self.eps = eps
        self.patience = patience

    def select(self, coverage):
        check_is_fitted(self)
        # round
        # self.M_ = np.round(self.M_).astype(int)
        self.pairs_with_incomplete_cover_ = np.argwhere(
            self.M_.sum(axis=1) < coverage).flatten()
        self.n_pairs_with_incomplete_cover_ = self.pairs_with_incomplete_cover_.size
        if self.n_pairs_with_incomplete_cover_ > 0:
            logger.info(
                f"Cannot cover {self.n_pairs_with_incomplete_cover_} elements.")
        self.coverage = coverage
        self.feature_importances_ = np.zeros(self.n_features_)

        # initial probabilities
        v_hat = np.full_like(self.feature_importances_, 1/2)
        no_change_iter = 0

        self.coverages_ = []
        self.average_n_features_selected_ = []
        ordered_features_ = []

        for iter in range(self.max_iters):
            # draw random samples
            samples = np.random.binomial(1, v_hat, (self.rs, self.n_features_)).astype(float)
            # compute scores for each sample
            scores = self._score(samples, self.alpha)
            # find quantile
            gamma_hat = np.quantile(scores, 1 - self.rho, interpolation='higher')
            prev_vhat = v_hat.copy()
            # update v-hat
            v_hat = ((scores >= gamma_hat)[:, np.newaxis] * samples).sum(axis=0)
            v_hat /= (scores >= gamma_hat).sum()
            # smoothed update
            if self.smoothing_parameter is not None:
                v_hat = (
                    self.smoothing_parameter * v_hat +
                    (1 - self.smoothing_parameter) * prev_vhat
                )
            # check if converged
            # logger.info(v_hat.max())
            # logger.info(np.abs(v_hat - prev_vhat).max())
            if np.abs(v_hat - prev_vhat).max() <= self.eps:
                no_change_iter += 1
                if no_change_iter == self.patience:
                    logger.info(f"Converged with eps={self.eps}.")
                    break
            else:
                no_change_iter = 0

            new_added = (
                set(np.argwhere(v_hat > 0.98).flatten().tolist()) -
                set(ordered_features_)
            )
            if len(new_added) > 0:
                ordered_features_ += list(new_added)
            high_prob_feats = np.argwhere(v_hat > 0.98).flatten()
            # logger.info(f"{high_prob_feats.size} high probability features.")
            # logger.info(f"{np.argwhere(v_hat > 0.5).flatten()}")
            per_element_coverage = self.M_[:, high_prob_feats].sum(axis=1)
            # logger.info(f"{per_element_coverage.min()} smallest coverage.")

        high_prob_feats = np.argwhere(v_hat > 0.98).flatten()
        assert set(high_prob_feats).issubset(set(ordered_features_))
        ordered_features_ = [i for i in ordered_features_
                                  if i in high_prob_feats]

        self.feature_importances_[ordered_features_] = np.arange(
            len(ordered_features_) + 1, 1, -1)

        logger.info(f"{high_prob_feats.size} high probability features.")
        per_element_coverage = self.M_[:, high_prob_feats].sum(axis=1)
        self.min_coverage = per_element_coverage.min()
        logger.info(f"{self.min_coverage} smallest coverage.")
        logger.info(f"{per_element_coverage.mean()} average coverage.")
        self.coverages_ = np.asarray(self.coverages_)
        self.solution = np.asarray(ordered_features_)
        self.average_n_features_selected_ = np.asarray(self.average_n_features_selected_)

        return self

    def _score(self, samples, alpha=None):
        """Score function for given random samples.

        samples: array-like of shape (N, n_features)
        alpha: hyperparameter determining trade-off between coverage
            and number of genes picked.

        Returns:
        scores: np.ndarray of shape (N,)
        """
        element_by_sample = self.M_ @ samples.T
        # Clip to desired coverage
        element_by_sample = np.clip(element_by_sample, 0, self.coverage)
        coverage_per_sample = element_by_sample.min(axis=0)
        genes_picked_per_sample = samples.sum(axis=1)
        if alpha is None:
            alpha = self.coverage / self.n_features_
        self.coverages_.append(coverage_per_sample.mean())
        self.average_n_features_selected_.append(genes_picked_per_sample.mean())
        return coverage_per_sample - alpha * genes_picked_per_sample



import os
from setuptools import setup, Command, find_packages


class CleanCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.system('rm -vrf ./build ./dist ./*.pyc ./*.tgz ./*.egg-info')


cmdclass = {'clean': CleanCommand}


options = {
    'name': 'phenotype_cover',
    'description': 'phenotype_cover is a package for biomarker discovery using multiset multicover.',
    'long_description': 'Implements the greedy and cross-entropy-method phenotype cover algorithms.',
    'license': 'MIT',
    'version': '0.2',
    'author': 'Euxhen Hasanaj',
    'author_email': 'ehasanaj@cs.cmu.edu',
    'url': 'https://github.com/euxhenh/phenotype-cover',
    'provides': ['phenotype_cover'],
    'package_dir': {'phenotype_cover': 'src/phenotype_cover'},
    'packages': find_packages(where='src'),
    'cmdclass': cmdclass,
    'platforms': 'ALL',
    'keywords': ['biomarker', 'marker', 'phenotype', 'scRNA-seq', 'set', 'cover', 'multiset', 'multicover'],
    'install_requires': ['numpy', 'matplotlib', 'scikit-learn', 'multiset-multicover'],
    'python_requires': ">=3.7"
}

setup(**options)
